/*  CPU_SPEED for GP2X
    Copyright (C) 2005  Hermes/PS2Reality 
	Modified by Vimacs and god_at_hell(LCD-part)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include "hw_gp2x.h"

void hwctrl_init()
{
	MEM_REG = &gp2x_memregs[0];
}

void load_system_regs()
{
	// Set UPLLSETVREG to 0x4F02, which gives 80MHz
	MEM_REG[0x0914>>1] = 0x4F02;

	// Wait for clock change to start
	while (MEM_REG[0x0902>>1] & 2);

	// Wait for clock change to be verified
	while (MEM_REG[0x0916>>1] != 0x4F02);
}

void set_FCLK(unsigned MHZ)
{
	unsigned v;
	unsigned mdiv,pdiv=3,scale=0;
	MHZ *= 1000000;
	mdiv = (MHZ*pdiv) / SYS_CLK_FREQ;
	mdiv = ((mdiv-8)<<8) & 0xff00;
	pdiv = ((pdiv-2)<<2) & 0xfc;
	scale &= 3;
	v = mdiv | pdiv | scale;
	MEM_REG[0x910>>1] = v;
}

unsigned get_freq_920_CLK()
{
	unsigned i;
	unsigned reg, mdiv, pdiv, scale;
	reg = MEM_REG[0x912>>1];
	mdiv = ((reg & 0xff00) >> 8) + 8;
	pdiv = ((reg & 0xfc) >> 2) + 2;
	scale = reg & 3;
	MDIV = mdiv;
	PDIV = pdiv;
	SCALE = scale;
	i = (MEM_REG[0x91c>>1] & 7)+1;
	return ((SYS_CLK_FREQ * mdiv)/(pdiv << scale))/i;
}

unsigned get_freq_940_CLK()
{
	unsigned i;
	unsigned reg, mdiv, pdiv, scale;
	reg=MEM_REG[0x912>>1];
	mdiv = ((reg & 0xff00) >> 8) + 8;
	pdiv = ((reg & 0xfc) >> 2) + 2;
	scale = reg & 3;
	i = ((MEM_REG[0x91c>>1]>>3) & 7)+1;
	return ((SYS_CLK_FREQ * mdiv) / (pdiv << scale))/i;
}

unsigned short get_920_Div()
{
	return (MEM_REG[0x91c>>1] & 0x7); 
}

unsigned short get_940_Div()
{
	return ((MEM_REG[0x91c>>1] >> 3) & 0x7); 
}

unsigned get_state940()
{
	return MEM_REG[0x904>>1];
}

unsigned short Disable_Int_940()
{
	unsigned short ret;
	ret = MEM_REG[0x3B42>>1];
	MEM_REG[0x3B42>>1] = 0;
	MEM_REG[0x3B46>>1] = 0xffff;	
	return ret;	
}

void Disable_940()
{
	Disable_Int_940();
	MEM_REG[0x3B48>>1] |= (1 << 7);
	MEM_REG[0x904>>1] &=0xfffe;
}

void Load_940_code(unsigned *code,int size)
{
	unsigned *cp;
	int i;
	arm940code = (unsigned short *)mmap(0, 0x100000, PROT_READ|PROT_WRITE, MAP_SHARED, gp2x_dev, 0x03000000);
	Disable_940();
	cp = (unsigned *) code;
	for (i = 0; i < size/4; i ++)
	{
		arm940code[i] = cp[i];
	}
	for (i = 0; i < 64; i ++)
	{
		arm940code[0x3FC0+i] = 0;
	}
	MEM_REG[0x3B48>>1] = (MEM_REG[0x3B48>>1]  & 0xFF00) | 0x03;
}

///////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////

void OpenMem()
{
	// Open /dev/mem
	if(!gp2x_dev) gp2x_dev = open("/dev/mem", O_RDWR);
	gp2x_memregs=(unsigned short *)mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, gp2x_dev, 0xc0000000);

	// Init CPU control
	hwctrl_init();
}

void OpenBatt()
{
	// Open /dev/batt
	if(!devbatt) devbatt = open ("/dev/batt", O_RDONLY);
}

int Get920Clock()
{
	unsigned sysfreq = 0;
	int cpufreq;

	OpenMem();

	// Get 920t clock
	printf("Getting 920t clock...\n");
	sysfreq = get_freq_920_CLK();
	sysfreq *= get_920_Div() + 1;
	cpufreq = sysfreq / 1000000;

	// Return clock frequency
	return cpufreq;
}

void Set920Clock(int Mhz)
{
	OpenMem();

	// Set 920t clock
	printf("Setting 920t clock to %i Mhz...\n", Mhz);
	set_FCLK(Mhz);
}

int Get940State()
{
	int i;
	char c[16];

	OpenMem();

	// Get 940t state
	sprintf(c, "%i", get_state940());
	if(c == "45055")
		i = 1;
	else
		i = 0;

	// Return state
	return i;
}

void Disable940()
{
	OpenMem();

	// Disable ARM 940t
	Disable_940();
}

void Enable940()
{
	OpenMem();

	// Enable ARM 940t
	MEM_REG[0x904>>1] |= 1;
}

void BatteryLED()
{
	OpenMem();

	// On/Off battery LED
	MEM_REG[GPIOHOUT >> 1] ^= 16;
}

void ShutdownScreen()
{
	OpenMem();

	// On/Off screen
	MEM_REG[GPIOHOUT >> 1] ^= 4;
}

int BatteryLEDState()
{
	int i;

	OpenMem();

	// Get battery LED state
	if(MEM_REG[GPIOHPINLVL >> 1] & 16)
		i = 0;
	else
		i = 1;

	// Return state
	return i;
}

int ScreenState()
{
	int i;

	OpenMem();

	// Get screen state
	if(MEM_REG[GPIOHPINLVL >> 1] & 4)
		i = 0;
	else
		i = 1;

	// Return state
	return i;
}

int GetBatteryCharge()
{
	int i = -1;
	unsigned short currentval = 0;

	OpenBatt();

	// Get battery charge
	read (devbatt, &currentval, 2);

	// Battery empty
	if(currentval < BATTERY_LO) i = 0; else
	// Battery medium
	if(currentval > BATTERY_LO && currentval < BATTERY_HI) i = 1; else
	// Battery full
	if(currentval > BATTERY_HI) i = 2; else
	// AC Adapter
	if(currentval > BATTERY_AC) i = 3;

	// Return i
	return i;
}
